/*  --- Copyright University of Sussex 1993. All rights reserved. ----------
 > File:           C.all/plog/lib/cd.pl
 > Purpose:        change directory (from PROLOG)
 > Author:         Jonathan Laventhol, Sep 7 1983 (see revisions)
 > Documentation:  HELP * CD
 > Related Files:
 */

:- format(";;; library(cd) is no longer needed~n").

/* --- Revision History ---------------------------------------------------
--- Simon Nichols, Aug 19 1993
		Changed warning message to standard form.
--- Simon Nichols, Jul 18 1990
		Obsolete: now a built-in command.
 */
