@NEWMASTER@ robd Tue Jul 17 18:07:15 BST 1990
214,219d
176c
define write_index(dir);
.
167c
define build_index(fs);
.
136c
define index_file(f, is);
.
126c
define add_to_index(ids, locn);
.
122c
		syssort(index(i), false, procedure(x, y);
			lvars x, y;
			alphabefore(front(x), front(y));
		endprocedure) -> index(i);
.
119c
define finish_index();
.
82,115c
define start_index();
.
77c
define is_blank(l);
.
70c
define is_entry_text(l);
.
50c
define is_entry_header(l);
.
30,46c
vars
.
28c
];
.
18,20c
constant keywords = [
.
4c
 * Author:          Rob Duncan, Mar 15 1990
.
@NEWMASTER@ robd Wed Aug 29 10:48:57 BST 1990
216,218d
171c
		index_file(f, instream(f));
.
@NEWMASTER@ robd Wed Oct 3 10:57:56 BST 1990
222,224d
190c
		cucharout(termin);
.
188c
			printf('%p %p %p %p %p\n', [% front(j), explode(back(j)) %]);
.
186c
		discout(dir dir_>< consstring(i,1)) -> cucharout;
.
177,184c
	lvars	i, j, dir;
	dlocal	cucharout, pop_pr_quotes = false, pop_file_versions = false;
.
@NEWMASTER@ robd Fri Mar 22 11:40:30 GMT 1991
223,225d
211c
	applist(ml_arglist, procedure(dir);
.
23d
1c
(* --- Copyright University of Sussex 1990. All rights reserved. ----------
.
@NEWMASTER@ robd Tue Jun 18 15:00:07 BST 1991
223,225d
214,215c
	applist(poparglist, procedure(dir);
		lvars dir;
		'Making doc_index for ' sys_>< dir =>
		mkindex(dir);
	endprocedure);
.
29,31d
@NEWMASTER@ robd Tue Jun 18 15:03:17 BST 1991
95c
			;;; (makes types take precedence over values)
.
@NEWMASTER@ robd Mon Aug 5 17:44:04 BST 1991
228,229d
215c
	write_index(dir dir_>< 'doc_index');
.
212c
			returnunless(issubstring('index', f))(dir dir_>< f -> f);
.
203,209c
	sys_file_match('*#', dir dir_>< '*', false, true) -> p;
	returnif(p() == termin);
	-> ; ;;; discard first pathname
	build_index(pdtolist(procedure() -> f;
.
16,17d
@NEWMASTER@ robd Mon Aug 5 18:14:29 BST 1991
204c
#_IF DEF VMS
	sys_file_match('*.;0', dir dir_>< '*', false, false) -> p;
#_ELSE
	sys_file_match('*#', dir dir_>< '*', false, false) -> p;
#_ENDIF
.
201d
