@NEWMASTER@ simonn Fri Jan 25 14:11:59 GMT 1991
36,41d
17,20c
string([]).
.
4c
 *  Author:         Jonathan Laventhol, July  4 1983
.
1,2c
/*  --- University of Sussex POPLOG file -----------------------------------
 *  File:           $usepop/master/C.all/plog/lib/strings.pl
.
