@rem --- Copyright University of Sussex 1997. All rights reserved. ----------
@rem File:			C.windows/extern/lib/mklibpop.bat
@rem Purpose:		Make Poplog external libraries
@rem Author:		Robert Duncan, May 29 1996 (see revisions)
@echo off

rem Must be run from $usepop/pop/extern/lib
cd ..\src
nmake -a nodebug=1 install
cd ..\lib

@rem --- Revision History ---------------------------------------------------
@rem --- Robert Duncan, Feb 28 1997 - now does full, optimised compile
