@NEWMASTER@ robd Fri Nov 2 15:30:34 GMT 1990
45,49d
15c
   vars n;
.
1,6c
/*  --- University of Sussex POPLOG file -----------------------------------
 *  File:           $usepop/master/C.all/plog/auto/are.pl
 *  Purpose:        generalisation of IS to handle more than one result
 *  Author:         Chris Mellish, July 1983
 *  Documentation:  HELP * PLOGTOPOP
 *  Related Files:
.
@NEWMASTER@ simonn Thu Sep 19 15:40:24 BST 1991
51,53d
44c
endprocedure -> prolog_valof("are",2);
.
41,42c
procedure x,y,contn;
.
37d
31,32c
   else x
.
29c
		 x.prolog_predword.valof.apply
.
27c
			prolog_eval_expr(prolog_arg(n,x))
.
23,24c
		 conspair(prolog_eval_expr(prolog_arg(1,x)),prolog_eval_expr(prolog_arg(2,x)))
.
15c
   lvars n;
.
1c
/*  --- Copyright University of Sussex 1990. All rights reserved. ----------
.
@NEWMASTER@ simonn Thu Nov 28 11:58:35 GMT 1991
41c
procedure();
.
