@NEWMASTER@ robd Fri Apr 20 14:47:28 BST 1990
32,33d
26c
fun isempty NilStack = true
.
23c
fun top (Push(x, _)) = x
.
20c
fun pop (Push s) = s
.
18c
val push = Push
.
16c
val nilstack = NilStack
.
10,11c
	NilStack
|	Push of 'a * 'a stack
.
1c
(* --- Copyright University of Sussex 1989.  All rights reserved. ---------
.
