@NEWMASTER@ robd Mon Jul 9 11:44:55 BST 1990
@NEWMASTER@ robd Mon Jul 16 12:17:02 BST 1990
82,92d
71,80d
55,64d
47,51d
45c
ml_val <	: fastint * fastint -> fastint = nonop fi_<;
ml_val <=	: fastint * fastint -> fastint = nonop fi_<=;
ml_val >	: fastint * fastint -> fastint = nonop fi_>;
ml_val >=	: fastint * fastint -> fastint = nonop fi_>=;
.
37c
ml_val ==	: fastint * fastint -> bool = nonop ==;
.
4c
 * Author:          Robert John Duncan, Jul  9 1990
.
@NEWMASTER@ robd Fri Aug 10 11:58:12 BST 1990
118,120d
45,72d
43a
ml_val div	: fastint * fastint -> fastint = nonop fi_div;
ml_val mod	: fastint * fastint -> fastint = nonop fi_rem;
.
39c
ml_val sign : fastint -> int;
.
13,15c
loadlib("int_parameters");
.
@NEWMASTER@ robd Tue Oct 30 12:02:10 GMT 1990
121,122d
73,75d
