;;; --- Copyright University of Sussex 1994. All rights reserved. ----------
;;; File:			C.windows/extern/src/poplog.def
;;; Purpose:		Module definition file for POPLOG.DLL
;;; Author:			Robert John Duncan, Jul  7 1994

LIBRARY poplog

EXPORTS
	pop_init_ast
	pop_init_heap
	pop_init_environment
	pop_init_registry
	pop_start

	;;; public callbacks
	pop_mishap
	pop_call
	pop_get_ident
	pop_free_fixed_hold
	pop_check_interrupt
	pop_sys_init
	pop_exfunc_arg DATA
	pop_external_flags DATA
