@NEWMASTER@ C.all/plog/lib/chartype.pl robd Thu Jan 12 12:06:53 GMT 1989
38,47c
	prolog_eval(apply(NewType, Char, updater(valof(prolog_chartype)))).
.
34c
	prolog_eval(prolog_chartype(Char), Type).
.
15a
:- prolog_language(pop11).

constant procedure
	prolog_chartype = item_chartype(% $-prolog$-prolog_standard_repeater %);

:- prolog_language(prolog).

.
4c
 > Author:        Robert Duncan, Simon Nichols, Feb. 1987
.
2c
 > File:          $usepop/master/C.all/plog/lib/chartype.pl
.
