@NEWMASTER@ C.all/plog/lib/dec10.pl robd Thu Sep  8 17:52:44 BST 1988
141,145d
21,26d
1,6c
/*  --- University of Sussex POPLOG file -----------------------------------
 *  File:           $usepop/master/C.all/plog/lib/dec10.pl
 *  Purpose:        DEC10 -- compatibility predicates
 *  Author:         Jonathan Laventhol, Mar 26 1984
 *  Documentation:  HELP * DEC10
 *  Related Files:
.
@NEWMASTER@ robd Mon Aug 21 14:48:08 BST 1989
154,155d
152d
78,81d
75d
40,41d
27,29d
1c
/*  --- Copyright University of Sussex 1988. All rights reserved. ----------
.
@NEWMASTER@ simonn Tue Mar 27 14:21:47 BST 1990
84,85d
12d
10a
:- op(1200, xfx, :-).
:- op(1200, xfx, -->).

:- op(1200, fx, :-).
:- op(1200, fx, ?-).

:- op(1150, fx, mode).
:- op(1150, fx, public).
:- op(1150, fx, dynamic).
:- op(1150, fx, static).
:- op(1150, fx, system_predicate).
:- op(1150, fx, user_predicate).
:- op(1150, fx, no_clauses).
:- op(1150, fx, clauses).
:- op(1150, fx, import).
:- op(1150, fx, export).
:- op(1150, fx, global).

:- op(1100, xfy, ;).

:- op(1050, xfy, ->).

:- op(1000, xfy, ',').

:- op(900, fy, \+).
:- op(900, fy, spy).
:- op(900, fy, nospy).
:- op(900, fx, module).
:- op(900, fx, endmodule).

:- op(700, xfx, =).
:- op(700, xfx, is).
:- op(700, xfx, =..).
:- op(700, xfx, ==).
:- op(700, xfx, \==).
:- op(700, xfx, @<).
:- op(700, xfx, @>).
:- op(700, xfx, @=<).
:- op(700, xfx, @>=).
:- op(700, xfx, =:=).
:- op(700, xfx, =\=).
:- op(700, xfx, <).
:- op(700, xfx, >).
:- op(700, xfx, =<).
:- op(700, xfx, >=).

:- op(500, yfx, +).
:- op(500, yfx, -).
:- op(500, yfx, /\).
:- op(500, yfx, \/).

:- op(500, fx, +).
:- op(500, fx, -).
:- op(500, fx, \).

:- op(400, yfx, *).
:- op(400, yfx, /).
:- op(400, yfx, <<).
:- op(400, yfx, >>).

:- op(300, xfx, div).
:- op(300, xfx, mod).

:- op(200, xfy, '^').

:- op(100, xfy, $-).
:- op(100, fy,  $-).


.
8a
;;; DEC-10 Prolog operator precedences

.
6c
 >  Related Files:
.
1c
/*  --- Copyright University of Sussex 1989. All rights reserved. ----------
.
@NEWMASTER@ simonn Tue Oct 8 17:31:10 BST 1991
83,85d
62a
	:- notimplemented(close(X)).
.
1c
/*  --- Copyright University of Sussex 1990. All rights reserved. ----------
.
@NEWMASTER@ simonn Tue Feb 11 12:57:15 GMT 1992
86,87c
		Removed dummy for definition for close/1 as there is now a
		sensible one in the Prolog system.
.
82,84d
73a
	leash(_) :- leash.
.
1c
/*  --- Copyright University of Sussex 1991. All rights reserved. ----------
.
