@NEWMASTER@ simonn Fri Mar 22 16:57:06 GMT 1991
43,45d
16c
        module(Goal,Prelist,Gains,Losses,Proc,Procargs),
.
1,2c
/*  --- University of Sussex POPLOG file -----------------------------------
 *  File:           $usepop/master/C.all/plog/lib/planet.pl
.
