@NEWMASTER@ andreass Mon Jul 23 15:17:12 BST 1990
31d
24c
extern WidgetClass popCompositeWidgetClass;
.
22c
#define XtNpopCallback "popCallback"
.
@NEWMASTER@ jonm Thu May 30 02:34:30 BST 1991
23,24d
6c
 * Documentation:
.
3c
 * Version:         Revision 3.0
.
@NEWMASTER@ jonm Thu Jun 27 16:45:47 BST 1991
35d
25,26d
1c
/* --- Copyright University of Sussex 1990. All rights reserved. ----------
.
@NEWMASTER@ johng Thu Oct 31 13:20:20 GMT 1991
28c
extern WidgetClass xpwCompositeWidgetClass;
.
