@NEWMASTER@ C.all/plog/lib/chatty.pl simonn Mon Dec 14 14:08:23 GMT 1987
180,181d
8,10c
:- current_op(Prec, xfx, ':-'),
    op(Prec, fx, chatty),
    op(Prec, fx, unchatty).
.
1,5c
/*  --- University of Sussex POPLOG file -----------------------------------
 *  File:           $usepop/master/C.all/plog/lib/chatty.pl
 *  Purpose:        make prolog clauses give a commentary on their execution.
 *  Author:         Roger Evans, May 1983 (see revisions)
 *  Documentation:  HELP * CHATTY
 *  Related Files:
.
