@NEWMASTER@ robd Fri Apr 20 14:43:22 BST 1990
43,44d
36c
fun isempty (Queue([], [])) = true
.
33,34c
|	front (Queue([], tail)) =
		front (Queue(rev(tail), []))
.
31c
|	front (Queue([], [])) =
.
29c
fun front (Queue(item :: head, tail)) =
.
26,27c
|	dequeue (Queue([], tail)) =
		dequeue (Queue(rev(tail), []))
.
22,24c
fun dequeue (Queue(item :: head, tail)) =
		(item, Queue(head, tail))
|	dequeue (Queue([], [])) =
.
19,20c
fun enqueue (item, Queue(head, tail)) =
		Queue(head, item :: tail)
.
17c
val nilqueue = Queue([], [])
.
11c
	Queue of 'a list * 'a list
.
1c
(* --- Copyright University of Sussex 1989.  All rights reserved. ---------
.
