@NEWMASTER@ ianr Thu Feb 23 17:19:36 GMT 1989
452,456d
256,257c
							1)) == NULL)
.
4c
 * Author:		Ben Rubinstein, Feb 20 1987
.
1,2c
/* --- Copyright University of Sussex 1987.  All rights reserved. ---------
 * File:		$usepop/master/C.sun/pwm/pwsetup.c
.
@NEWMASTER@ ianr Tue Feb 27 13:54:30 GMT 1990
464,466d
253,259d
184,185d
139d
1c
/* --- Copyright University of Sussex 1989.  All rights reserved. ---------
.
@NEWMASTER@ johnw Thu Jan 31 17:45:50 GMT 1991
465,466d
463d
258c
	/*** Colour-pwm ved/base-win text speedup ***/
.
255c
	if (BASE || com_charargs[1] == 'v')
.
253c
	/*** Colour-pwm ved/base-win text speedup ***/
.
1c
/* --- Copyright University of Sussex 1990.  All rights reserved. ---------
.
