@NEWMASTER@ C.all/plog/lib/useful.pl robd Tue Feb  9 14:43:47 GMT 1988
17,45c
/* load a library file unless already loaded */
    uses(Pred) :-
        current_predicate(Pred, X), !.
    uses(Pred) :-
.
15a
/* load(Filename) ;;; '.pl' not needed */
    load(X) :-
        Y is consword(><(X, '.pl')),
        reconsult(Y).
.
12,14c
/* once(Goal) */
    once(X) :-
        call(X), !.
.
9,10c
/* append(Left, Right, Leftright) */
    append([], L, L).
    append([H|R], S, [H|T]) :-
        append(R, S, T).
.
7a
/* member(Element, List) */
    member(X, [X|_]).
    member(X, [_|Y]) :-
        member(X, Y).
.
2,5c
 *  File:           $usepop/master/C.all/plog/lib/useful.pl
 *  Purpose:        useful predicates: member/2, append/3, once/1, load/1
 *  Author:         Unknown, 6 July 1983
 *  Documentation:  HELP * USEFUL
 *  Related Files:
.
@NEWMASTER@ robd Mon Apr 9 12:44:58 BST 1990
54,59d
15d
8,13d
4c
 >  Author:         Unknown
.
1c
/*  --- University of Sussex POPLOG file -----------------------------------
.
