@rem --- Copyright University of Sussex 1996. All rights reserved. ----------
@rem File:			C.windows/pop/pgcomp.bat
@rem Purpose:		Compile POPLOG source files (.p or .s)
@rem Author:		Robert Duncan, May 29 1996
@echo off

if "%1" == "-cs" goto :cs
corepop +popc %1 %2 %3 %4 %5 %6 %7 %8 %9
goto end
:cs
corepop +poplibr -lcs ./ > P_FILES
if errorlevel 1 goto end
corepop +popc -l @P_FILES
if errorlevel 1 goto end
dir /b *.w > W_FILES
:end
if exist P_FILES del P_FILES
