/* --- Copyright University of Sussex 1991.  All rights reserved. ---------
 * File:        C.pwm/pwm/pwid.h
 * Purpose:     defines parts of PWM identity message
 * Author:      Ben Rubinstein, Feb 20 1987 (see revisions)
 */


/*
|   This file defines parts of the "PWM long identity" message, which
|   may differ according to machine and PWM version.
|
|   Those which are commented out must be supplied elsewhere in the
|   linking process.  Note that screen width/height are already declared in
|   PWDEC.H
|
*/

#define PWMID_version       "1.80"
#define Version_str         "Sussex POPLOG Window Manager, version 1.80:"
#define Version_len         43          /* length of string above */

/* #define PWMID_date          "2/10/86" */     /* supplied by makefile */
/* #define PWMID_machine       "sun3" */        /* supplied by makefile */

#ifdef DeBug
#define PWMID_misc           "debug"
#else
#define PWMID_misc           "-"
#endif


/* --- Revision History ---------------------------------------------------
--- John Williams, Feb  7 1991
		Updated version to 1.80 (cf. BR isl.27)
 */
