@NEWMASTER@ andreass Fri Jul 20 16:24:31 BST 1990
@NEWMASTER@ andreass Mon Jul 23 15:18:15 BST 1990
60d
55c
extern XpwCompositeClassRec popCompositeClassRec;
.
51c
   XpwCompositeClassPart popComposite_class;
.
41c
   XpwCompositePart popcomposite;
.
34c
	XtCallbackList popCallback;
.
18c
 *  3. new fields for the class popCom
.
16c
 *  1. new fields for instances of the popCom - a structure.
.
10c
 * Related Files:   XpwCom.h XpwCom.c popgraphics.p
.
@NEWMASTER@ jonm Thu May 30 02:33:30 BST 1991
66,67d
40d
26,33c
#include <CompositeP.h>
#include <XpwComposit.h>
.
16,19c
 *  1. new fields for instances of the xpwCom - a structure.
 *  2. declaration of the full instance record
 *  3. new fields for the class xpwCom
 *  4. declaration of the full class record.
 *  ps - these naming conventions aren't mine ... and yes, they are horrid.
.
13d
11c

.
8,9c
 * Related Files:   XpwCom.h XpwCom.c xpwgraphics.p
.
3,4c
 * Version:         Revision 3.0
 * Project:         widgetise pop
 * Location:        <XpwCompositeP.h>
 * Purpose:         this is the private header file for the widget.
.
@NEWMASTER@ jonm Thu Jun 27 16:45:09 BST 1991
41d
@NEWMASTER@ johng Mon Oct 28 15:06:25 GMT 1991
33d
30d
@NEWMASTER@ johng Thu Oct 31 13:21:01 GMT 1991
64c
extern XpwCompositeClassRec xpwCompositeClassRec;
.
