@NEWMASTER@ johnw Tue Apr 24 12:30:03 BST 1990
176,182d
128c
    environ[i] = "TERM=pwm";                /* replace it in the array */
.
4c
 * Author:		Ben Rubinstein, Jan  8 1987
.
1,2c
/* --- Copyright University of Sussex 1987.  All rights reserved. ---------
 * File:		$usepop/master/C.sun/pwm/pwpty.c
.
