@NEWMASTER@ C.all/plog/lib/popsetq.pl robd Tue Feb  9 14:42:30 GMT 1988
11,19c
/* --- Revision History ---------------------------------------------------
--- Kathryn Seifert, Sep 10 1986  Changed to export the procedure
    'prolog_val' rather than 'prolog_valof'.
*/
.
9c
procedure;
lvars this_section;
    current_section -> this_section;
    prolog_section -> current_section;
    section_export("prolog_setq");
    section_export("psetq");
    section_export("prolog_val");
    this_section -> current_section;
endprocedure();
.
7c
:- prolog_language(pop11).
.
5a
;;; to make  ?- prolog_setq(Pop_variable_name, Value)
;;; and      ?- prolog_val(Pop_variable_name, Value)
;;; work.
.
2,3c
 *  File:           $usepop/master/C.all/plog/lib/popsetq.pl
 *  Purpose:        for setting the values of pop variables from prolog
 *  Author:         Unknown, ??? (see revisions)
 *  Documentation:  HELP * POPSETQ
 *  Related Files:
.
