@NEWMASTER@ johnw Thu Feb 7 16:30:22 GMT 1991
30,35d
18,19c
#define PWMID_version       "1.78"
#define Version_str         "Sussex POPLOG Window Manager, version 1.78:"
.
4c
 * Author:      Ben Rubinstein, Feb 20 1987
.
1,2c
/* --- Copyright University of Sussex 1987.  All rights reserved. ---------
 * File:        $usepop/master/C.sun/pwm/pwid.h
.
