@NEWMASTER@ ianr Tue Jan 17 14:30:06 GMT 1989
710,715d
437,441d
429,435c
			co_insert_char('^');
			co_insert_char(code + 64);
			co_carriage_return();
			send_to_poplog(26);
.
4c
 * Author:  	Ben Rubinstein, Jan  8 1987
.
2c
 * File:		$usepop/master/C.sun/pwm/pwinput.c
.
@NEWMASTER@ ianr Wed Jan 25 17:02:54 GMT 1989
701a

/* --- Revision History ---------------------------------------------------
--- Ian Rogers, Jan 17 1989
	--- changed behaviour of ctrl-Z in base window (this should quit pwmtool)
		such that it now sends the string 'bye' instead ascii 26
 */
.
433a
			echo_code_and_save('b');
			echo_code_and_save('y');
			echo_code_and_save('e');
			dispatch_input_line(0);

.
429,432c
		/* The old quit code: IR
		*
		*	co_insert_char('^');
		*	co_insert_char(code + 64);
		*	co_carriage_return();
		*	send_to_poplog(26);
		*/
.
4c
 * Author:  	Ben Rubinstein, Jan  8 1987 (see revisions)
.
2c
 * File:		C.sun/pwm/pwinput.c
.
@NEWMASTER@ ianr Wed Jan 25 17:04:32 GMT 1989
4c
 * Author:  	Ben Rubinstein, Jan  8 1987
.
1,2c
/* --- Copyright University of Sussex 1987.  All rights reserved. ---------
 * File:		$usepop/master/C.sun/pwm/pwinput.c
.
@NEWMASTER@ johnw Tue Jul 24 16:58:54 BST 1990
722,727d
284c
				wmgr_setupmenu(toolfd);
				mi = menu_display(&wmgr_toolmenu, &inevent, toolfd);
.
276,282d
248d
20,33c
extern struct menu *wmgr_toolmenu;
.
1c
/* --- Copyright University of Sussex 1989.  All rights reserved. ---------
.
