@NEWMASTER@ ianr Wed Feb 14 00:39:21 GMT 1990
26,27d
16c
PATH=$popsys\:$PATH\:$popcom             
.
2c
# --- Copyright University of Sussex 1989. All rights reserved. ----------
.
@NEWMASTER@ simonn Fri Oct 5 12:31:52 BST 1990
26,27d
16c
PATH=$popsys\:$Xpopbin\:$PATH\:$popcom
.
