/* --- Copyright University of Sussex 1996. All rights reserved. ----------
 * File:			C.win32/extern/src/main.c
 * Purpose:			Main routine for Poplog as a console application
 * Author:			Robert John Duncan, Jun 14 1994 (see revisions)
 * Documentation:
 * Related Files:
 */

#include "popcore.h"

void main(int argc, char *argv[]) {
	UINT exitCode;
	ExitProcess(pop_start(TRUE, &exitCode) ? exitCode : 1);
}

/* --- Revision History ---------------------------------------------------
--- Robert John Duncan, Feb  7 1996
		Common startup code moved to pop_start.
 */
