@NEWMASTER@ ianr Tue Jan 17 14:36:39 GMT 1989
272,277d
86c
		blop = ((code == LOC_WINEXIT) ? PIX_CLR : PIX_SET);
.
4c
 * Author:		Ben Rubinstein, Feb 20 1987
.
2c
 * File:		$usepop/master/C.sun/pwm/pwboil.c
.
