/* --- Copyright University of Sussex 1992. All rights reserved. ----------
 > File:           C.all/plog/lib/findall.pl
 > Purpose:		   findall/3: finds all solutions to a goal
 > Author:         Robert Duncan, March 10 1989 (see revisions)
 > Documentation:  HELP * FINDALL
 */

:- format(";;; library(findall) is no longer needed~n").

/* --- Revision History ---------------------------------------------------
--- Robert John Duncan, Apr 10 1992
		Moved findall/3 into the system: see C.all/plog/src/sortpreds.p
--- Rob Duncan, Apr  9 1990
		Added global declaration.
--- Rob Duncan, Mar 10 1989
		Completely rewritten in POP-11
 */
