@NEWMASTER@ aarons Wed Feb 6 19:43:44 GMT 1991
638,643d
470,471d
468d
457,466d
452d
446,447c
*	moving the mouse into it  (it would be neater to check, after opening
*	the window, whether the mouse is now within it, and if so avoid moving
*	it: but this is surprisingly laborious.  Perhaps in the next version?)
.
4c
 * Author:		Ben Rubinstein, Jan  8 1987
.
1,2c
/* --- Copyright University of Sussex 1987.  All rights reserved. ---------
 * File:		$usepop/master/C.sun/pwm/pwcommand.c
.
