pop_pop11="-$popsavelib/startup.psv"; export pop_pop11
pop_prolog="$pop_pop11 -$popsavelib/prolog.psv"; export pop_prolog
pop_clisp="$pop_pop11 -$popsavelib/clisp.psv"; export pop_clisp
pop_pml="$pop_pop11 +$popsavelib/pml.psv"; export pop_pml
pop_xved="$pop_pop11 +$popsavelib/xved.psv"; export pop_xved
pop_ved="$pop_pop11 :sysinitcomp();ved"; export pop_ved
pop_help="$pop_pop11 :sysinitcomp();help"; export pop_help
pop_ref="$pop_pop11 :sysinitcomp();ref"; export pop_ref
pop_teach="$pop_pop11 :sysinitcomp();teach"; export pop_teach
pop_doc="$pop_pop11 :sysinitcomp();doc"; export pop_doc
pop_im="$pop_pop11 :sysinitcomp();im"; export pop_im
