@NEWMASTER@ C.sun/pwm/pwdec.h johnw Fri Jun  5 16:27:45 GMT 1987
557,558d
38,39d
@NEWMASTER@ C.sun/pwm/pwdec.h aledm Tue Nov 10 18:46:29 GMT 1987
558,560d
41,43c
#define SCREENWIDTH     1152
#define SCREENHEIGHT    900
.
2c
 * File:           $usepop/master/C.sun/pwm/pwdec.h
.
