@NEWMASTER@ andreass Mon Jul 23 15:19:57 BST 1990
58d
53c
extern WidgetClass popCoreWidgetClass;
.
42c
#define XtNpopCallback  "popCallback"
.
30c
 popCallback     Callback      Pointer      NULL
.
@NEWMASTER@ jonm Thu May 30 03:49:18 BST 1991
60,61d
37d
17,22c
#include <Intrinsic.h>  /* for XtSpecificationRelease */
#include <XpwMethod.h>      /* methods for XpwCore widgets + descendants */
.
6a
 * Notes:           The Core widget is just a blank widget with callbacks,
 *                  and a few other resources. It is the top of the Poplog
 *                  widget hierarchy.
.
3,4c
 * Version:         Revision 3.0
 * Purpose:         this is the public header file for the widget.
.
@NEWMASTER@ jonm Thu May 30 03:49:42 BST 1991
47a
#define XtNFid "Fid"
.
36a
 tileForeground	 Foreground    Pixel	    XtDefaultForeground
.
@NEWMASTER@ johng Fri Aug 2 03:57:40 BST 1991
60,61d
48,49d
1c
/* --- Copyright University of Sussex 1990. All rights reserved. ----------
.
@NEWMASTER@ johng Thu Oct 31 13:22:27 GMT 1991
55c
extern WidgetClass xpwCoreWidgetClass;
.
