@NEWMASTER@ C.all/plog/lib/lispinplog.pl johnw Thu Sep  3 17:59:22 GMT 1987
100,104d
77,89d
4c
 > Author:         John Williams, Apr 30 1987
.
@NEWMASTER@ C.all/plog/lib/lispinplog.pl johnw Fri Mar  4 16:45:50 GMT 1988
105,106d
103d
96,97d
2c
 > File:           $usepop/master/C.all/plog/lib/lispinplog.pl
.
@NEWMASTER@ johnw Wed Jul 18 18:32:46 BST 1990
115,116d
108d
105,106d
103a
nonmac lisp -> prolog_macro("LISP");
.
97d
76,78d
64d
54d
43d
36d
1c
/* --- Copyright University of Sussex 1987.  All rights reserved. ---------
.
@NEWMASTER@ johnw Wed Jul 18 18:38:33 BST 1990
79,81d
