/* --- Copyright University of Sussex 1996. All rights reserved. ----------
 * File:			C.win32/extern/src/winmain.c
 * Purpose:			Main routine for Poplog as a GUI application
 * Author:			Robert John Duncan, Jul 15 1994 (see revisions)
 * Documentation:
 * Related Files:
 */

#include "popcore.h"

int APIENTRY WinMain(
		HINSTANCE	hInstance,
		HINSTANCE	hPrevInstance,
		LPSTR		lpszCmdLine,
		int			nCmdShow
	)
{
	UINT exitCode;
	return pop_start(TRUE, &exitCode) ? exitCode : 1;
}

/* --- Revision History ---------------------------------------------------
--- Robert John Duncan, Feb  7 1996
		Common startup code moved to pop_start.
 */
