/* --- Copyright University of Sussex 1996. All rights reserved. ----------
 * File:			C.win32/extern/src/x.c
 * Purpose:			Utilities needed when linking with X in NuTCRACKER
 * Author:			Robert Duncan, May 29 1996 (see revisions)
 * Related Files:	Original versions of these functions are in c_core.c
 */

#include "popcore.h"
#include "../lib/c_core.h"


/*
 *		Struct used by _pop_set_xt_wakeup
 */
globaldef XTWAKE __pop_xt_wakeup_struct;

void _pop_do_interrupt()
{
}

int _pop_sigmask(int block)
{
	return 0;
}


/* --- Revision History ---------------------------------------------------
--- Robert Duncan, Oct 28 1996
		Moved polling code out to new file "pop_poll.c"
 */
