/* --- Copyright University of Sussex 1996. All rights reserved. ----------
 * File:			C.win32/extern/src/dllentry.c
 * Purpose:			Default entry point for DLLs not using CRT
 * Author:			Robert Duncan, Jun 19 1996
 */

#include <windows.h>

BOOL WINAPI _DllMainCRTStartup(HINSTANCE hDLL, DWORD reason, void* reserved) {
	if (reason == DLL_PROCESS_ATTACH)
		/* Don't call again for each thread attach/detach */
		DisableThreadLibraryCalls(hDLL);
	return TRUE;
}
