/* --- Copyright University of Sussex 1994. All rights reserved. ----------
 * File:			C.win32/extern/src/test.c
 * Purpose:			Test harness for functions in popcore.lib
 * Author:			Robert John Duncan, Aug  3 1994
 */

#include "popcore.h"

DWORD _pop_signals_pending = 0;

void __pop_errsig(void)
{
	DWORD *psc = (DWORD*)&__pop_sigcontext;
	fprintf(stderr, "<<<<<< Exception - 0x%.8x >>>>>>\n", psc[1]);
	ExitProcess(1);
}

int pop_main(void)
{
	return 0;
}
