@NEWMASTER@ C.sun/pwm/pwmain.c aledm Tue Nov 10 18:46:41 GMT 1987
215,220d
77c
		mishap(-1, "PWM: couldn't make base window (outside Suntools???)\n");
.
44,53c
	rootfd = open("/dev/win0", 0);
.
40,42d
36d
22,24d
7,10d
4c
 * Author:         Ben Rubinstein, Feb 20 1987
.
2c
 * File:           $usepop/master/C.sun/pwm/pwmain.c
.
