@NEWMASTER@ C.sun/pwm/pwgfxtext.c ianr Tue Jun  7 15:23:07 BST 1988
142,145d
128,129c
						graphic_op, cg_pixfontp, com_stringarg);
.
123,124c
						graphic_op, cg_pixfontp, com_stringarg);
.
4c
 * Author:  	Ben Rubinstein, Feb 20 1987
.
@NEWMASTER@ C.sun/pwm/pwgfxtext.c pop Wed Aug 17 15:24:02 BST 1988
144d
128c
						graphic_op | PIX_COLOUR(graphic_value), cg_pixfontp,
.
123c
						graphic_op | PIX_COLOUR(graphic_value), cg_pixfontp,
.
@NEWMASTER@ ianr Thu Feb 23 17:45:25 GMT 1989
147d
17a

.
10,16c
#define CSUN_TXT_BUG    /* colour-sun bug: pw_text screws up on pixels of    *
						* depth>1; so use pw_ttext on gfx windows, and don't *
						** use >2 colours on txt windows.                    */
.
1,2c
/* --- Copyright University of Sussex 1987.  All rights reserved. ---------
 * File:        $usepop/master/C.sun/pwm/pwgfxtext.c
.
