@NEWMASTER@ ianr Thu Feb 23 17:40:28 GMT 1989
538,542d
325,328c
					mem_create(newrect.r_width, newrect.r_height, 1))
				== NULL)
.
271,273c
								newrect.r_height + fontadv_y, 1))
.
4c
 * Author:		Ben Rubinstein, Jan  8 1987
.
1,2c
/* --- Copyright University of Sussex 1987.  All rights reserved. ---------
 * File:        $usepop/master/C.sun/pwm/pwmisc.c
.
