@NEWMASTER@ andreass Mon Jul 23 15:23:42 BST 1990
74d
69c
extern XpwGraphicClassRec popGraphicClassRec;
.
63,65c
   XpwCoreClassPart popcore_class;
   XpwPixmapClassPart poppixmap_class;
   XpwGraphicClassPart popgraphic_class;
.
52,54c
   XpwCorePart popcore;
   XpwPixmapPart poppixmap;
   XpwGraphicPart popgraphic;
.
17c
 *  3. new fields for the class popGraphic
.
15c
 *  1. new fields for instances of the popGraphic - a structure.
.
9c
 * Related Files:   XpwGraphic.h XpwGraphic.c popgraphics.p
.
@NEWMASTER@ jonm Thu May 30 14:31:36 BST 1991
81,82d
53d
50c
	 char *key;
.
26,33c
#include <XpwGraphic.h>
#include <XpwPixmapP.h>
.
22a

.
17,20c
 *  1. new fields for instances of the xpwGraphic - a structure.
 *  2. declaration of the full instance record
 *  3. new fields for the class xpwGraphic
 *  4. declaration of the full class record.
 *  ps - these naming conventions aren't mine ... and yes, they are horrid.
.
14d
12d
9d
3,4c
 * Version:         Revision 3.0
 * Project:         widgetise pop
 * Purpose:         this is the private header file for the widget.
.
@NEWMASTER@ johng Thu Oct 31 13:23:38 GMT 1991
76c
extern XpwGraphicClassRec xpwGraphicClassRec;
.
@NEWMASTER@ jonm Tue Dec 17 13:41:22 GMT 1991
80,81d
52a
	 Vmap *vmap;
.
1c
/* --- Copyright University of Sussex 1990. All rights reserved. ----------
.
