@NEWMASTER@ ianr Thu Feb 23 17:51:53 GMT 1989
933,937d
424a
		for (msiz = 1; i >> msiz != 1 ; msiz++) {};
	}

.
423a
		i = gfx_mapsizes[(wt_colmap[current_graf])];
.
422a
	else
	{
		int i;
.
408,421c
	if ((cg_winisframe) || (dpth_r == 1))
.
4c
 * Author:  Ben Rubinstein (+ bitconversion by Aled Morris), Jan  8 1987
.
1,2c
/* --- Copyright University of Sussex 1987.  All rights reserved. ---------
 * File:	$usepop/master/C.sun/pwm/pwgraster.c
.
@NEWMASTER@ simonn Mon Aug 12 15:06:05 BST 1991
941,943d
737d
731,734d
11d
1c
/* --- Copyright University of Sussex 1989.  All rights reserved. ---------
.
