@rem --- Copyright University of Sussex 1996. All rights reserved. ----------
@rem File:			C.windows/pop/pglibr.bat
@rem Purpose:		Update POPLOG module library
@rem Author:		Robert Duncan, May 29 1996
@echo off

if "%1" == "" goto noargs
corepop +poplibr %1 %2 %3 %4 %5 %6 %7 %8 %9
goto end
:noargs
corepop +poplibr -r ./ @W_FILES
if errorlevel 1 goto end
del W_FILES *.w *.o
:end
