@rem --- Copyright University of Sussex 1997. All rights reserved. ----------
@rem File:			C.windows/extern/unicode/mkencodings.bat
@rem Purpose:		Build encoding DLLs
@rem Author:		Robert Duncan, Mar  6 1997
@rem Documentation:
@rem Related Files:
@echo off

for %%i in ( *.c ) do call :mkdll %%~ni
copy *.dll ..\lib
del *.dll *.obj *.exp *.lib
goto :EOF

:mkdll
cl /nologo /c /O2 %1.c
link /nologo /dll /out:%1.dll /export:%1,DATA %1.obj
