@NEWMASTER@ andreass Mon Jul 23 15:26:46 BST 1990
63d
57c
extern XpwPixmapClassRec popPixmapClassRec;
.
52,53c
   XpwCoreClassPart popcore_class;
   XpwPixmapClassPart poppixmap_class;
.
42,43c
   XpwCorePart popcore;
   XpwPixmapPart poppixmap;
.
17c
 *  3. new fields for the class popPixmap
.
15c
 *  1. new fields for instances of the popPixmap - a structure.
.
9c
 * Related Files:   XpwPixmap.h XpwPixmap.c popgraphics.p
.
@NEWMASTER@ jonm Thu May 30 14:30:30 BST 1991
69,70d
26,33c
#include <XpwPixmap.h>
#include <XpwCoreP.h>
.
22a

.
17,20c
 *  1. new fields for instances of the xpwPixmap - a structure.
 *  2. declaration of the full instance record
 *  3. new fields for the class xpwPixmap
 *  4. declaration of the full class record.
 *  ps - these naming conventions aren't mine ... and yes, they are horrid.
.
14d
12d
9d
6,7c
 * Date:            11 February 1981
 * Documentation:
.
3,4c
 * Version:         Revision 3.0
 * Project:         widgetise pop
 * Purpose:         this is the private header file for the widget.
.
@NEWMASTER@ johng Thu Oct 31 13:15:46 GMT 1991
75,79d
63c
extern XpwPixmapClassRec xpwPixmapClassRec;
.
1c
/* --- Copyright University of Sussex 1990. All rights reserved. ----------
.
