@NEWMASTER@ C.sun/pwm/pwparse.c johnw Fri Jun  5 16:28:32 GMT 1987
401,404d
304,305c
		if (++com_stringlen >= ARGSTRSIZE) com_stringlen--;
		com_stringarg[com_stringlen] = c;
.
4c
 * Author:		Ben Rubinstein, Feb 20 1987
.
