@NEWMASTER@ C.sun/pwm/pwimages.h rogere Mon Nov  9 12:06:29 GMT 1987
80,81c
/* this is basically "/usr/include/images/painting_hand.pr". I can't use it
*   directly, because it's not wide enough.  BUt I'm sure Sun won't mind...
*   this is it with bordered white area round.
.
