/* --- Copyright University of Sussex 1994. All rights reserved. ----------
 * File:			C.vms/extern/lib/pop_shrim_start.cc
 * Purpose:			Define dummy procedure __pop_shrim_start
 * Author:			John Gibson, May  6 1993 (see revisions)
 */

/*	A dummy identifier to mark where shareable images start (this file is put
 *	last in the link, and the linker puts the code section last, so
 *	it must be a procedure).
 *
 *	(This file has extension .cc and object file has extension .object
 *	so it doesn't get incorporated in libpop.olb.)
 */

#ifdef __alpha
/* demand-zero part of default cluster gets put nearer the end, so use an
 * uninitialised array
 */
int __pop_shrim_start[4];
#else
void __pop_shrim_start() {;}
#endif


/* --- Revision History ---------------------------------------------------
--- John Gibson, Nov 21 1994
		Added Alpha version
 */
