/*  --- Copyright University of Sussex 1988.  All rights reserved. ---------
 >  File:			C.all/plog/lib/default_ops.pl
 >  Purpose:        Restore the default operator precedences
 >  Author:         Robert Duncan, March 1988
 >  Documentation:  HELP * DEFAULT_OPS
 */


/*
 *	The default precedences are simply those of the PDP-11,
 *	but it simplifies documentation not to have to say that
 */

:- library(pdp11).
