@NEWMASTER@ robd Fri Aug 10 12:13:05 BST 1990
74,78d
61c
ml_val sign : ratio -> int;
.
48c
ml_val floor : ratio -> int = ml_valof('StdValues.floor');
.
45d
26a
	lconstant Ratio = ml_valof("Ratio");
.
25d
4c
 * Author:          Robert John Duncan, Jul 13 1990
.
